package de.ozgcloud.xta.client.core;

import java.util.Optional;

import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.InvalidMessageIDException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.MessageSchemaViolationException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.MessageVirusDetectionException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.ParameterIsNotValidException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.PermissionDeniedException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.SyncAsyncException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;
import de.ozgcloud.xta.client.model.XtaIdentifier;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.Builder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

@Builder
@RequiredArgsConstructor
@Slf4j
public class XtaClientService {

	private final WrappedXtaService service;
	private final XtaClientConfig config;

	static final String TRANSPORT_REPORT_FAILED_ERROR = "Failed to get transport report!";

	public Optional<XtaTransportReport> getTransportReport(XtaMessageMetaData messageMetaData) {
		try {
			return Optional.of(getTransportReportOrThrowException(messageMetaData));
		} catch (XtaClientRuntimeException e) {
			logError(TRANSPORT_REPORT_FAILED_ERROR, e);
			return Optional.empty();
		}
	}

	public void closeMessage(XtaMessage message) {
		var messageId = message.metaData().messageId();
		var readerClientIdentifier = message.metaData().readerIdentifier();
		try {
			service.close(messageId, readerClientIdentifier);
		} catch (XTAWSTechnicalProblemException | PermissionDeniedException | InvalidMessageIDException | RuntimeException e) {
			logError("Failed to close message! '%s' (reader: %s)".formatted(messageId, readerClientIdentifier), e);
		}
	}

	public Optional<XtaMessage> getMessage(XtaMessageMetaData messageMetaData) {
		var messageId = messageMetaData.messageId();
		var readerClientIdentifier = messageMetaData.readerIdentifier();
		try {
			return Optional.of(service.getMessage(messageId, readerClientIdentifier));
		} catch (XTAWSTechnicalProblemException | PermissionDeniedException | InvalidMessageIDException | RuntimeException e) {
			logError("Failed to get message by id! '%s' (reader: %s)".formatted(messageId, readerClientIdentifier.value()), e);
			return Optional.empty();
		}
	}

	public Optional<XtaMessageMetaDataListing> getStatusList(XtaIdentifier clientIdentifier) {
		try {
			return Optional.of(service.getStatusList(clientIdentifier, config.getMaxListItems()));
		} catch (PermissionDeniedException | XTAWSTechnicalProblemException | RuntimeException e) {
			logError("Failed to get status list!", e);
			return Optional.empty();
		}
	}

	void logError(String message, Exception e) {
		log.error(message, e);
	}

	public boolean checkAccountActive(XtaIdentifier clientIdentifier) {
		try {
			service.checkAccountActive(clientIdentifier);
			return true;
		} catch (XTAWSTechnicalProblemException | RuntimeException e) {
			throw new XtaClientRuntimeException("Failed to check account active!", e);
		} catch (PermissionDeniedException e) {
			return false;
		}
	}

	public XtaTransportReport sendMessage(XtaMessage messageWithoutMessageId) {
		var message = getXtaMessageWithMessageId(messageWithoutMessageId);
		try {
			service.sendMessage(message);
			return getTransportReportOrThrowException(message.metaData());
		} catch (XTAWSTechnicalProblemException | PermissionDeniedException | SyncAsyncException | ParameterIsNotValidException |
				MessageVirusDetectionException | MessageSchemaViolationException | RuntimeException e) {
			throw new XtaClientRuntimeException("Failed to send message!", e);
		}
	}

	XtaMessage getXtaMessageWithMessageId(XtaMessage messageWithoutMessageId) {
		var authorIdentifier = messageWithoutMessageId.metaData().authorIdentifier();
		try {
			var messageId = service.createMessageId(authorIdentifier);
			return createXtaMessageWithMessageId(messageWithoutMessageId, messageId);
		} catch (XTAWSTechnicalProblemException | PermissionDeniedException | RuntimeException e) {
			throw new XtaClientRuntimeException("Failed to create message id!", e);
		}
	}

	XtaMessage createXtaMessageWithMessageId(XtaMessage messageWithoutMessageId, String messageId) {
		return messageWithoutMessageId.toBuilder()
				.metaData(messageWithoutMessageId.metaData().toBuilder()
						.messageId(messageId)
						.build())
				.build();
	}

	XtaTransportReport getTransportReportOrThrowException(XtaMessageMetaData messageMetaData) {
		var messageId = messageMetaData.messageId();
		var authorId = messageMetaData.authorIdentifier();
		try {
			return service.getTransportReport(messageId, authorId);
		} catch (XTAWSTechnicalProblemException | PermissionDeniedException | InvalidMessageIDException | RuntimeException e) {
			throw new XtaClientRuntimeException(
					"Failed to get transport report! (messageId: %s, reader: %s)".formatted(messageId, authorId.value()), e);
		}
	}

	public boolean lookupService(XtaMessageMetaData messageMetaData) {
		try {
			return service.lookupService(
					messageMetaData.service(),
					messageMetaData.readerIdentifier(),
					messageMetaData.authorIdentifier()
			);
		} catch (XTAWSTechnicalProblemException | PermissionDeniedException | ParameterIsNotValidException | RuntimeException exception) {
			logWarning("Failed to lookup service for message '%s'!".formatted(messageMetaData.messageId()), exception);
			return false;
		}
	}

	void logWarning(String message, Exception exception) {
		log.warn(message, exception);
	}

}
