package de.ozgcloud.xta.client.core;

import java.util.List;
import java.util.Map;

import jakarta.xml.ws.BindingProvider;

import org.apache.cxf.endpoint.Client;
import org.apache.cxf.ext.logging.LoggingInInterceptor;
import org.apache.cxf.ext.logging.LoggingOutInterceptor;
import org.apache.cxf.frontend.ClientProxy;
import org.apache.cxf.interceptor.Interceptor;
import org.apache.cxf.message.Message;
import org.apache.cxf.transport.http.HTTPConduit;

import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.exception.XtaClientInitializationException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.XTAService;
import lombok.Builder;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

@Builder
@RequiredArgsConstructor
@Slf4j
public class XtaPortTripleFactory {

	static final String SCHEMA_VALIDATION_ENABLED_KEY = "schema-validation-enabled";

	private final XTAService xtaService;
	private final XtaClientConfig config;
	private final XtaTLSClientParametersFactory tlsClientParametersFactory;

	public static XtaPortTripleFactory from(final XtaClientConfig config) {
		return XtaPortTripleFactory.builder()
				.config(config)
				.xtaService(new XTAService())
				.tlsClientParametersFactory(XtaTLSClientParametersFactory.builder()
						.config(config)
						.build())
				.build();
	}

	public XtaPortTriple create() throws XtaClientInitializationException {
		log.debug("[createXtaService] Using config: {}", config);
		return new XtaPortTriple(
				configurePort(config.getManagementServiceUrl(), xtaService.getManagementPort()),
				configurePort(config.getSendServiceUrl(), xtaService.getSendXtaPort()),
				configurePort(config.getMsgBoxServiceUrl(), xtaService.getMsgBoxPort())
		);
	}

	private <T> T configurePort(final String endpointUrl, final T port) throws XtaClientInitializationException {
		var bindingProvider = (BindingProvider) port;
		configureBinding(endpointUrl, bindingProvider);
		return port;
	}

	void configureBinding(final String endpointUrl, final BindingProvider port) throws XtaClientInitializationException {
		configureRequestContext(endpointUrl, port.getRequestContext());
		configureClient(getClientFromPort(port));
	}

	Client getClientFromPort(BindingProvider port) {
		return ClientProxy.getClient(port);
	}

	void configureRequestContext(final String endpointUrl, Map<String, Object> requestContext) {
		requestContext.putAll(Map.of(
				BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpointUrl,
				SCHEMA_VALIDATION_ENABLED_KEY, config.isSchemaValidation(),
				"com.sun.xml.ws.request.timeout", 10000,
				"com.sun.xml.ws.connect.timeout", 10000
		));
	}

	void configureClient(Client client) throws XtaClientInitializationException {
		if (config.isLogSoapRequests()) {
			configureOutInterceptors(client.getOutInterceptors());
		}
		if (config.isLogSoapResponses()) {
			configureInInterceptors(client.getInInterceptors());
		}
		configureHttpConduit((HTTPConduit) client.getConduit());
		log.debug("[configureClient] Initialized TransportSecurity");
	}

	void configureInInterceptors(List<Interceptor<? extends Message>> inInterceptors) {
		inInterceptors.add(createInInterceptor());
	}

	void configureOutInterceptors(List<Interceptor<? extends Message>> outInterceptors) {
		outInterceptors.add(createOutInterceptor());
	}

	private LoggingOutInterceptor createOutInterceptor() {
		var outInterceptor = new LoggingOutInterceptor();
		outInterceptor.setPrettyLogging(true);
		return outInterceptor;
	}

	private LoggingInInterceptor createInInterceptor() {
		var inInterceptor = new LoggingInInterceptor();
		inInterceptor.setPrettyLogging(true);
		return inInterceptor;
	}

	void configureHttpConduit(HTTPConduit conduit) throws XtaClientInitializationException {
		conduit.setTlsClientParameters(tlsClientParametersFactory.create());
	}
}
