package de.ozgcloud.xta.client.model;

import java.math.BigInteger;
import java.time.ZonedDateTime;

import jakarta.annotation.Nullable;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotBlank;
import jakarta.validation.constraints.NotNull;
import jakarta.validation.constraints.PositiveOrZero;

import lombok.Builder;

/**
 * Represents the metadata of an XTA message.
 *
 * <p>
 * <small>Parameter documentation strings partially from <i>OSCI_MessageMetaData_V2.02.xsd</i>.</small>
 * </p>
 *
 * @param service Distinct service in a certain business scenario context; in the XÖV context this is the "Dienste URI" (service URI). <p>This field is required.</p>
 * @param businessScenarioCode Domain qualifier, e.g. Meldewesen, XVergabe... <p>This field is required.</p>
 * @param businessScenarioListUri Code list URI of the business scenario. <p>If {@code null}, {@link XtaMessageMetaData#businessScenarioCode} refers to an undefined business scenario. Otherwise, it is a defined scenario which additionally requires {@link XtaMessageMetaData#businessScenarioListVersionId}.</p>
 * @param businessScenarioListVersionId code list version ID of the business scenario. <p>Required if {@link XtaMessageMetaData#businessScenarioListUri} is not {@code null}.</p>
 * @param messageTypeCode Code of the message type. <p>For instance, <i>Geschaeftsgang.Geschaeftsgang.0201</i> with business scenario <i>XDOMEAGAD_DATA</i>.</p> <p>This field is required.</p>
 * @param messageTypePayloadSchema Payload schema of the message type. <p>For instance, <i>urn:xoev-de:xdomea:schema:2.4.0</i> with code <i>Geschaeftsgang.Geschaeftsgang.0201</i> and business scenario <i>XDOMEAGAD_DATA</i>.</p> <p>This field is required.</p>
 * @param messageId Unique identifier of the message. <p>The ID is generated by the server and thus not required when sending.</p>
 * @param authorIdentifier Identifier of the author of the message. <p>This field is required.</p>
 * @param readerIdentifier Identifier of the reader of the message. <p>This field is required.</p>
 * @param messageSize Size of the message in bytes. <p>The size is determined by the server and thus not required when sending.</p>
 * @param deliveryAttributesOrigin Timestamp when the message was sent. <p>This field is only set for received messages and not required when sending.</p>
 * @param deliveryAttributesDelivery Timestamp when the message was delivered. <p>This field is only set for received messages and not required when sending.</p>
 */
@Builder(toBuilder = true)
public record XtaMessageMetaData(
		@NotBlank String service,
		@NotBlank String businessScenarioCode,
		@Nullable String businessScenarioListUri,
		@Nullable String businessScenarioListVersionId,
		@NotBlank String messageTypeCode,
		@NotBlank String messageTypePayloadSchema,
		@Nullable String messageId,
		@NotNull @Valid XtaIdentifier authorIdentifier,
		@NotNull @Valid XtaIdentifier readerIdentifier,
		@Nullable @PositiveOrZero BigInteger messageSize,
		@Nullable ZonedDateTime deliveryAttributesOrigin,
		@Nullable ZonedDateTime deliveryAttributesDelivery
) {
}
