package de.ozgcloud.xta.client.xdomea;

import static de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper.*;

import java.util.List;
import java.util.Set;

import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaXmlValues;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReader;
import lombok.Builder;
import lombok.RequiredArgsConstructor;

@Builder
@RequiredArgsConstructor
public class XdomeaMetaDataValidator {

	private final ZipFileEntryReader zipFileEntryReader;

	static final String AUTHOR_ID_PREFIX_NAME = "author";
	static final String READER_ID_PREFIX_NAME = "reader";

	public void validate(
			XtaFile xdomeaZipFile,
			XdomeaXmlValues xdomeaXmlValues,
			XtaMessageMetaData messageMetadataData
	) throws XtaClientException {
		validateIdentifierPrefixes(
				xdomeaXmlValues.authorIdPrefix(),
				xdomeaXmlValues.readerIdPrefix()
		);
		validateZipFileName(
				xdomeaZipFile.name(),
				xdomeaXmlValues.processId(),
				messageMetadataData.messageTypeCode()
		);
		validatePrimaryDocumentReferences(xdomeaZipFile, xdomeaXmlValues.primaryDocumentNames());
	}

	void validateIdentifierPrefixes(String authorIdPrefix, String readerIdPrefix) throws XtaClientException {
		validateIdentifierPrefix(authorIdPrefix, AUTHOR_ID_PREFIX, AUTHOR_ID_PREFIX_NAME);
		validateIdentifierPrefix(readerIdPrefix, READER_ID_PREFIX, READER_ID_PREFIX_NAME);
	}

	void validateIdentifierPrefix(String prefix, String expectedPrefix, String prefixName) throws XtaClientException {
		if (!expectedPrefix.equals(normalizePrefix(prefix))) {
			throw new XtaClientException("Expect prefix of %s identifier to be '%s'! (actual: %s)".formatted(prefixName, expectedPrefix, prefix));
		}
	}

	private String normalizePrefix(String prefix) {
		return prefix.endsWith(":") ? prefix.substring(0, prefix.length() - 1) : prefix;
	}

	void validatePrimaryDocumentReferences(XtaFile xdomeaZipFile, List<String> primaryDocumentNames) throws XtaClientException {
		var entryNames = Set.copyOf(zipFileEntryReader.getEntryNames(xdomeaZipFile.content()));
		for (var primaryDocumentName : primaryDocumentNames) {
			if (!entryNames.contains(primaryDocumentName)) {
				throw new XtaClientException("Primary document reference '%s' not found in xdomea zip file!".formatted(primaryDocumentName));
			}
		}
	}

	void validateZipFileName(String xdomeaZipFileName, String processId, String messageTypeCode) throws XtaClientException {
		var expectedXdomeaZipFileName = "%s_%s.zip".formatted(processId, messageTypeCode);
		if (!expectedXdomeaZipFileName.equals(xdomeaZipFileName)) {
			throw new XtaClientException(
					"Expect xdomea zip file name to equal '%s'! (actual: '%s')".formatted(expectedXdomeaZipFileName, xdomeaZipFileName)
			);
		}
	}

}
