package de.ozgcloud.xta.client.xdomea;

import static java.util.Collections.*;

import java.io.IOException;
import java.io.InputStream;

import javax.xml.parsers.DocumentBuilder;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaValueReader;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaXmlValues;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReader;
import lombok.Builder;
import lombok.RequiredArgsConstructor;

/**
 * Creator of {@link XtaMessage}s for Xdomea.
 *
 * <p>
 * The XdomeaXtaMessageCreator is used to create an {@link XtaMessage} for an Xdomea zip file. The resulting message may be send via the
 * {@link de.ozgcloud.xta.client.XtaClient XtaClient}.
 * </p>
 *
 *
 * Example:
 * <pre>
 * var xtaMessageCreator = XdomeaXtaMessageCreator.createInstance();
 * var zipFileName = "d5be7468-e620-4126-a40e-61a7f9b46345_Geschaeftsgang.Geschaeftsgang.0201.zip";
 * var zipFileContentDataHandler = new DataHandler(new FileDataSource("/path/to/" + zipFileName));
 * var xdomeaXtaFile = XtaFile.builder()
 *   .name(zipFileName)
 *   .content(zipFileContentDataHandler)
 *   .contentType("application/zip")
 *   .build();
 * var xtaMessage = xtaMessageCreator.createMessage(xdomeaXtaFile);
 *
 * XtaClient.from(...).sendMessage(xtaMessage);
 * </pre>
 */
@Builder
@RequiredArgsConstructor
public class XdomeaXtaMessageCreator {

	private final XdomeaValueReader xdomeaValueReader;
	private final DocumentBuilder documentBuilder;
	private final MetadataMapper metadataMapper;
	private final ZipFileEntryReader zipFileEntryReader;
	private final XdomeaMetaDataValidator metaDataValidator;

	public static XdomeaXtaMessageCreator createInstance() {
		return XdomeaXtaMessageCreatorFactory.createInstance().create();
	}

	/**
	 * Creates an {@link XtaMessage} from the given xdomea zip file.
	 *
	 * <p>
	 * The zip file has to follow the XTA/Xdomea generic transport specification. According to the specification, only the Xdomea messages of type
	 * 0201, 0401 and 0402 are supported.
	 * </p>
	 *
	 * <p>
	 * <small>The XTA/Xdomea generic transport specification is unofficially available as <i>"Transportfestlegungen Generischer
	 * Antragsdienst.pdf"</i>.</small>
	 * </p>
	 *
	 * @param xdomeaZipFile the xdomea zip file
	 * @return the xta message with metadata derived from the xdomea zip file
	 * @throws XtaClientException if no valid metadata could be derived from the xdomea zip file.
	 */
	public XtaMessage createMessage(XtaFile xdomeaZipFile) throws XtaClientException {
		return XtaMessage.builder()
				.metaData(deriveValidMetaData(xdomeaZipFile))
				.messageFile(xdomeaZipFile)
				.attachmentFiles(emptyList())
				.build();
	}

	XtaMessageMetaData deriveValidMetaData(XtaFile xdomeaZipFile) throws XtaClientException {
		try {
			var xdomeaXmlValues = readXdomeaXmlValues(xdomeaZipFile);
			var messageMetadataData = metadataMapper.mapXtaMessageMetadata(xdomeaXmlValues);
			metaDataValidator.validate(xdomeaZipFile, xdomeaXmlValues, messageMetadataData);
			return messageMetadataData;
		} catch (XtaClientRuntimeException e) {
			throw new XtaClientException("Failed to derive valid message metadata from xdomea document!", e);
		}
	}

	XdomeaXmlValues readXdomeaXmlValues(XtaFile xdomeaZipFile) throws XtaClientException {
		var document = readXdomeaXmlDocument(xdomeaZipFile);
		return xdomeaValueReader.readValues(document);
	}

	Document readXdomeaXmlDocument(XtaFile xdomeaZipFile) {
		return zipFileEntryReader.mapEntryData(
				xdomeaZipFile.content(),
				replaceFilenameExtensionWithXml(xdomeaZipFile.name()),
				this::parseXmlDocument
		);
	}

	private Document parseXmlDocument(InputStream inputStream) {
		try {
			return documentBuilder.parse(inputStream);
		} catch (SAXException | IOException e) {
			throw new XtaClientRuntimeException("Failed to parse xml document!", e);
		}
	}

	String replaceFilenameExtensionWithXml(String name) {
		return name.replaceFirst("\\.[^.]+$", ".xml");
	}

}
