package de.ozgcloud.xta.client.core;

import static de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import jakarta.xml.ws.Holder;

import org.apache.cxf.ws.addressing.AttributedURIType;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.xta.client.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.GenericContentContainer;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.LookupServiceRequest;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.LookupServiceResponse;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.ManagementPortType;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.MsgBoxPortType;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.SendPortType;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.TransportReport;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgBoxCloseRequestType;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgBoxFetchRequest;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgBoxStatusListRequestType;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgStatusListType;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2014.x10.transport.PartyType;
import de.ozgcloud.xta.client.mapper.RequestMapper;
import de.ozgcloud.xta.client.mapper.ResponseMapper;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.SneakyThrows;

class WrappedXtaServiceTest {

	@Mock
	private XtaPortTriple xtaPortTriple;

	@Mock
	private RequestMapper requestMapper;

	@Mock
	private ResponseMapper responseMapper;

	@InjectMocks
	private WrappedXtaService service;

	@DisplayName("get message")
	@Nested
	class TestGetMessage {

		@Mock
		private MsgBoxPortType msgBoxPortType;

		@Mock
		private MsgBoxFetchRequest msgBoxFetchRequest;

		@Mock
		private PartyType partyType;

		@Mock
		private MessageMetaData messageMetaData;

		@Mock
		private GenericContentContainer genericContentContainer;

		@Mock
		private XtaMessage xtaMessage;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xtaPortTriple.msgBoxPort()).thenReturn(msgBoxPortType);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(partyType);
			when(requestMapper.mapMsgBoxFetchRequest(MESSAGE_ID)).thenReturn(msgBoxFetchRequest);

			when(msgBoxPortType.getMessage(eq(msgBoxFetchRequest), eq(partyType), any(), eq(null))).thenAnswer(invocation -> {
				@SuppressWarnings("unchecked")
				var metaData = (Holder<MessageMetaData>) invocation.getArgument(2);
				metaData.value = messageMetaData;
				return genericContentContainer;
			});
			when(responseMapper.mapXtaMessage(genericContentContainer, messageMetaData))
					.thenReturn(xtaMessage);
		}

		@DisplayName("should return message")
		@Test
		@SneakyThrows
		void shouldReturnMessage() {
			var result = service.getMessage(MESSAGE_ID, SELF_IDENTIFIER);

			assertThat(result).isEqualTo(xtaMessage);
		}
	}

	@DisplayName("check account active")
	@Nested
	class TestCheckAccountActive {

		@Mock
		private PartyType partyType;

		@Mock
		private ManagementPortType managementPortType;

		@BeforeEach
		void mock() {
			when(xtaPortTriple.managementPort()).thenReturn(managementPortType);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(partyType);
		}

		@DisplayName("should call raw service method")
		@Test
		@SneakyThrows
		void shouldCallRawServiceMethod() {
			service.checkAccountActive(SELF_IDENTIFIER);

			verify(managementPortType).checkAccountActive(partyType);
		}
	}

	@DisplayName("get status list")
	@Nested
	class TestGetStatusList {
		@Mock
		private MsgBoxPortType msgBoxPortType;

		@Mock
		private MsgBoxStatusListRequestType msgBoxStatusListRequestType;

		@Mock
		private PartyType partyType;

		@Mock
		private MsgBoxResponseType msgBoxResponseType;

		@Mock
		private MsgStatusListType msgStatusListType;

		@Mock
		private XtaMessageMetaDataListing xtaMessageMetaDataListing;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xtaPortTriple.msgBoxPort()).thenReturn(msgBoxPortType);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(partyType);
			when(requestMapper.mapMsgBoxStatusListRequestTypeFromMaxListItems(MAX_LIST_ITEMS))
					.thenReturn(msgBoxStatusListRequestType);
			when(msgBoxPortType.getStatusList(eq(msgBoxStatusListRequestType), eq(partyType), any()))
					.thenAnswer(invocation -> {
						@SuppressWarnings("unchecked")
						var responseType = (Holder<MsgBoxResponseType>) invocation.getArgument(2);
						responseType.value = msgBoxResponseType;
						return msgStatusListType;
					});
			when(responseMapper.mapXtaMessageMetaListing(msgStatusListType, msgBoxResponseType))
					.thenReturn(xtaMessageMetaDataListing);
		}

		@DisplayName("should map response")
		@Test
		@SneakyThrows
		void shouldMapResponse() {
			var result = service.getStatusList(SELF_IDENTIFIER, MAX_LIST_ITEMS);

			assertThat(result).isEqualTo(xtaMessageMetaDataListing);
		}
	}

	@DisplayName("send message")
	@Nested
	class TestSendMessage {

		@Mock
		private SendPortType sendPortType;

		@Mock
		private GenericContentContainer genericContentContainer;

		@Mock
		private MessageMetaData messageMetaData;

		private XtaMessage xtaMessage;

		@BeforeEach
		void mock() {
			xtaMessage = XtaMessageTestFactory.create();
			when(xtaPortTriple.sendPort()).thenReturn(sendPortType);
			when(requestMapper.mapGenericContentContainer(xtaMessage)).thenReturn(genericContentContainer);
			when(requestMapper.mapMessageMetaDataFromXtaMessageMetaData(xtaMessage.metaData())).thenReturn(messageMetaData);
		}

		@DisplayName("should call port method")
		@Test
		@SneakyThrows
		void shouldPortMethod() {
			service.sendMessage(xtaMessage);

			verify(sendPortType).sendMessage(genericContentContainer, messageMetaData, null);
		}
	}

	@DisplayName("create message id")
	@Nested
	class TestCreateMessageId {

		@Mock
		private ManagementPortType managementPortType;

		@Mock
		private PartyType authorPartyType;

		@Mock
		private AttributedURIType messageIdType;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xtaPortTriple.managementPort()).thenReturn(managementPortType);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(authorPartyType);
			when(managementPortType.createMessageId(authorPartyType)).thenReturn(messageIdType);
			when(messageIdType.getValue()).thenReturn(MESSAGE_ID);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			var result = service.createMessageId(SELF_IDENTIFIER);

			assertThat(result).isEqualTo(MESSAGE_ID);
		}
	}

	@DisplayName("close")
	@Nested
	class TestClose {

		@Mock
		private MsgBoxPortType msgBoxPortType;

		@Mock
		private PartyType partyType;

		@Mock
		private MsgBoxCloseRequestType closeRequestType;

		@BeforeEach
		void mock() {
			when(xtaPortTriple.msgBoxPort()).thenReturn(msgBoxPortType);
			when(requestMapper.mapMsgBoxCloseRequestType(MESSAGE_ID)).thenReturn(closeRequestType);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(partyType);
		}

		@DisplayName("should call port method")
		@Test
		@SneakyThrows
		void shouldReturn() {
			service.close(MESSAGE_ID, SELF_IDENTIFIER);

			verify(msgBoxPortType).close(closeRequestType, partyType);
		}

	}

	@DisplayName("lookup service")
	@Nested
	class TestLookupService {

		@Mock
		private ManagementPortType managementPortType;

		@Mock
		private LookupServiceRequest lookupServiceRequest;

		@Mock
		private PartyType clientPartyType;

		@Mock
		private LookupServiceResponse lookupServiceResponse;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xtaPortTriple.managementPort()).thenReturn(managementPortType);
			when(requestMapper.mapLookupServiceRequest(MESSAGE_SERVICE, READER_IDENTIFIER)).thenReturn(lookupServiceRequest);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(clientPartyType);
			when(managementPortType.lookupService(lookupServiceRequest, clientPartyType)).thenReturn(lookupServiceResponse);
		}

		@DisplayName("should return is available")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		@SneakyThrows
		void shouldReturnIsAvailable(Boolean isAvailable) {
			when(responseMapper.isServiceAvailable(lookupServiceResponse)).thenReturn(isAvailable);

			var result = service.lookupService(MESSAGE_SERVICE, READER_IDENTIFIER, SELF_IDENTIFIER);

			assertThat(result).isEqualTo(isAvailable);

		}
	}

	@DisplayName("get transport report")
	@Nested
	class TestGetTransportReport {

		@Mock
		private ManagementPortType managementPortType;

		@Mock
		private AttributedURIType messageIdType;

		@Mock
		private PartyType authorPartyType;

		@Mock
		private TransportReport transportReport;

		@Mock
		private XtaTransportReport xtaTransportReport;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xtaPortTriple.managementPort()).thenReturn(managementPortType);
			when(requestMapper.mapAttributedURIType(MESSAGE_ID)).thenReturn(messageIdType);
			when(requestMapper.mapPartyTypeFromIdentifier(SELF_IDENTIFIER)).thenReturn(authorPartyType);
			when(managementPortType.getTransportReport(messageIdType, authorPartyType)).thenReturn(transportReport);
			when(responseMapper.mapXtaTransportReport(transportReport)).thenReturn(xtaTransportReport);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			var result = service.getTransportReport(MESSAGE_ID, SELF_IDENTIFIER);

			assertThat(result).isEqualTo(xtaTransportReport);
		}
	}

}