package de.ozgcloud.xta.client.core;

import static de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataListingTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageTestFactory;
import de.ozgcloud.xta.client.factory.XtaTransportReportTestFactory;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.PermissionDeniedException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.XTAWSTechnicalProblemException;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.model.XtaMessageMetaDataListing;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.SneakyThrows;

class XtaClientServiceTest {

	@Mock
	private WrappedXtaService wrapper;

	@Mock
	private XtaClientConfig config;

	@Spy
	@InjectMocks
	private XtaClientService service;

	private final XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
	private final XtaMessage message = XtaMessageTestFactory.create();
	private final XtaTransportReport transportReport = XtaTransportReportTestFactory.create();

	@DisplayName("get transport report")
	@Nested
	class TestGetTransportReport {

		@Mock
		private XtaClientRuntimeException exception;

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			doReturn(transportReport).when(service).getTransportReportOrThrowException(messageMetaData);

			var result = service.getTransportReport(messageMetaData);

			assertThat(result).contains(transportReport);
		}

		@DisplayName("should return empty if exception")
		@Test
		void shouldReturnEmptyIfException() {
			doThrow(exception).when(service).getTransportReportOrThrowException(messageMetaData);

			var result = service.getTransportReport(messageMetaData);

			assertThat(result).isEmpty();
		}

		@DisplayName("should log error if exception")
		@Test
		void shouldLogErrorIfException() {
			doThrow(exception).when(service).getTransportReportOrThrowException(messageMetaData);

			service.getTransportReport(messageMetaData);

			verify(service).logError(anyString(), eq(exception));
		}
	}

	@DisplayName("close message")
	@Nested
	class TestCloseMessage {
		@Mock
		private XTAWSTechnicalProblemException xtaWSTechnicalProblemException;

		@DisplayName("should call close")
		@Test
		@SneakyThrows
		void shouldCallClose() {
			service.closeMessage(message);

			verify(wrapper).close(MESSAGE_ID, READER_IDENTIFIER);
		}

		@DisplayName("should log error if exception")
		@Test
		@SneakyThrows
		void shouldLogErrorIfException() {
			doThrow(xtaWSTechnicalProblemException).when(wrapper).close(MESSAGE_ID, READER_IDENTIFIER);

			service.closeMessage(message);

			verify(service).logError(anyString(), eq(xtaWSTechnicalProblemException));
		}

	}

	@DisplayName("get message")
	@Nested
	class TestGetMessage {

		@Mock
		private XTAWSTechnicalProblemException xtaWSTechnicalProblemException;

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			doReturn(message).when(wrapper).getMessage(MESSAGE_ID, READER_IDENTIFIER);

			var result = service.getMessage(messageMetaData);

			assertThat(result).contains(message);
		}

		@DisplayName("should return empty if exception")
		@Test
		@SneakyThrows
		void shouldReturnEmptyIfException() {
			doThrow(xtaWSTechnicalProblemException).when(wrapper).getMessage(MESSAGE_ID, READER_IDENTIFIER);

			var result = service.getMessage(messageMetaData);

			assertThat(result).isEmpty();
		}

		@DisplayName("should log error if exception")
		@Test
		@SneakyThrows
		void shouldLogErrorIfException() {
			doThrow(xtaWSTechnicalProblemException).when(wrapper).getMessage(MESSAGE_ID, READER_IDENTIFIER);

			service.getMessage(messageMetaData);

			verify(service).logError(anyString(), eq(xtaWSTechnicalProblemException));
		}
	}

	@DisplayName("get status list")
	@Nested
	class TestGetStatusList {

		@Mock
		private XTAWSTechnicalProblemException xtaWSTechnicalProblemException;

		private final XtaMessageMetaDataListing listing = XtaMessageMetaDataListingTestFactory.create();

		@BeforeEach
		void mock() {
			when(config.getMaxListItems()).thenReturn(MAX_LIST_ITEMS);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			doReturn(listing).when(wrapper).getStatusList(SELF_IDENTIFIER, MAX_LIST_ITEMS);

			var result = service.getStatusList(SELF_IDENTIFIER);

			assertThat(result).contains(listing);
		}

		@DisplayName("should return empty if exception")
		@Test
		@SneakyThrows
		void shouldReturnEmptyIfException() {
			doThrow(xtaWSTechnicalProblemException).when(wrapper).getStatusList(SELF_IDENTIFIER, MAX_LIST_ITEMS);

			var result = service.getStatusList(SELF_IDENTIFIER);

			assertThat(result).isEmpty();
		}

		@DisplayName("should log error if exception")
		@Test
		@SneakyThrows
		void shouldLogErrorIfException() {
			doThrow(xtaWSTechnicalProblemException).when(wrapper).getStatusList(SELF_IDENTIFIER, MAX_LIST_ITEMS);

			service.getStatusList(SELF_IDENTIFIER);

			verify(service).logError(anyString(), eq(xtaWSTechnicalProblemException));
		}
	}

	@DisplayName("check account active")
	@Nested
	class TestCheckAccountActive {

		@DisplayName("should call check account active")
		@Test
		@SneakyThrows
		void shouldCallCheckAccountActive() {
			checkAccountActive();

			verify(wrapper).checkAccountActive(SELF_IDENTIFIER);
		}

		@DisplayName("should return true if no exception")
		@Test
		void shouldReturnTrueIfNoException() {
			var result = checkAccountActive();

			assertThat(result).isTrue();
		}

		@DisplayName("should return false if permission denied exception")
		@Test
		@SneakyThrows
		void shouldReturnFalseIfPermissionDeniedException() {
			doThrow(new PermissionDeniedException()).when(wrapper).checkAccountActive(any());

			var result = checkAccountActive();

			assertThat(result).isFalse();
		}

		@DisplayName("should throw on technical problem exception")
		@Test
		@SneakyThrows
		void shouldThrowOnTechnicalProblemException() {
			var technicalException = new XTAWSTechnicalProblemException();
			doThrow(technicalException).when(wrapper).checkAccountActive(any());

			assertThatThrownBy(this::checkAccountActive)
					.isInstanceOf(XtaClientRuntimeException.class)
					.hasCause(technicalException);
		}

		private boolean checkAccountActive() {
			return service.checkAccountActive(SELF_IDENTIFIER);
		}
	}

	@DisplayName("send message")
	@Nested
	class TestSendMessage {

		@Mock
		private XtaMessage messageWithoutMessageId;

		@Mock
		private XTAWSTechnicalProblemException exception;

		@BeforeEach
		@SneakyThrows
		void mock() {
			doReturn(message).when(service).getXtaMessageWithMessageId(messageWithoutMessageId);
		}

		@DisplayName("should call send message")
		@Test
		@SneakyThrows
		void shouldCallSendMessage() {
			sendMessage();

			verify(wrapper).sendMessage(message);
		}

		@DisplayName("should return with transport report")
		@Test
		@SneakyThrows
		void shouldReturnWithTransportReport() {
			doReturn(transportReport).when(service).getTransportReportOrThrowException(message.metaData());

			var result = sendMessage();

			assertThat(result).isEqualTo(transportReport);
		}

		@DisplayName("should throw client runtime exception if send message fails")
		@Test
		@SneakyThrows
		void shouldThrowClientRuntimeExceptionIfSendMessageFails() {
			doThrow(exception).when(wrapper).sendMessage(message);

			assertThatThrownBy(this::sendMessage)
					.isInstanceOf(XtaClientRuntimeException.class)
					.hasCause(exception);
		}

		private XtaTransportReport sendMessage() {
			return service.sendMessage(messageWithoutMessageId);
		}
	}

	@DisplayName("get xta message with message id")
	@Nested
	class TestGetXtaMessageWithMessageId {

		@Mock
		private XtaMessage messageWithMessageId;

		@Mock
		private XTAWSTechnicalProblemException exception;

		@DisplayName("should return message")
		@Test
		@SneakyThrows
		void shouldReturnMessage() {
			doReturn(MESSAGE_ID2).when(wrapper).createMessageId(AUTHOR_IDENTIFIER);
			doReturn(messageWithMessageId).when(service).createXtaMessageWithMessageId(message, MESSAGE_ID2);

			var result = service.getXtaMessageWithMessageId(message);

			assertThat(result).isEqualTo(messageWithMessageId);
		}

		@DisplayName("should throw client runtime exception if create message fails")
		@Test
		@SneakyThrows
		void shouldThrowClientRuntimeExceptionIfCreateMessageFails() {
			doThrow(exception).when(wrapper).createMessageId(AUTHOR_IDENTIFIER);

			assertThatThrownBy(() -> service.getXtaMessageWithMessageId(message))
					.isInstanceOf(XtaClientRuntimeException.class)
					.hasCause(exception);
		}
	}

	@DisplayName("get transport report or throw exception")
	@Nested
	class TestGetTransportReportOrThrowException {
		@Mock
		private XTAWSTechnicalProblemException exception;

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			when(wrapper.getTransportReport(MESSAGE_ID, AUTHOR_IDENTIFIER)).thenReturn(transportReport);

			var result = service.getTransportReportOrThrowException(messageMetaData);

			assertThat(result).isEqualTo(transportReport);
		}

		@DisplayName("should throw client exception if exception")
		@Test
		@SneakyThrows
		void shouldThrowClientExceptionIfException() {
			when(wrapper.getTransportReport(MESSAGE_ID, AUTHOR_IDENTIFIER)).thenThrow(exception);

			assertThatThrownBy(() -> service.getTransportReportOrThrowException(messageMetaData))
					.isInstanceOf(XtaClientRuntimeException.class)
					.hasCause(exception);
		}
	}

	@DisplayName("create xta message with message id")
	@Nested
	class TestCreateXtaMessageWithMessageId {
		private final XtaMessage xtaMessage = XtaMessageTestFactory.create();

		@DisplayName("should set message id")
		@Test
		void shouldSetMessageId() {
			var newMessageId = "newMessageId";

			var result = service.createXtaMessageWithMessageId(xtaMessage, newMessageId);

			assertThat(result.metaData().messageId()).isEqualTo(newMessageId);
		}
	}

	@DisplayName("lookup service")
	@Nested
	class TestLookupService {
		@Mock
		private XTAWSTechnicalProblemException exception;

		@DisplayName("should return lookup result")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		@SneakyThrows
		void shouldReturnLookupResult(boolean lookupResult) {
			when(wrapper.lookupService(MESSAGE_SERVICE, READER_IDENTIFIER, AUTHOR_IDENTIFIER)).thenReturn(lookupResult);

			var result = lookupService();

			assertThat(result).isEqualTo(lookupResult);
		}

		@DisplayName("should return false if exception")
		@Test
		@SneakyThrows
		void shouldReturnFalseIfException() {
			when(wrapper.lookupService(MESSAGE_SERVICE, READER_IDENTIFIER, AUTHOR_IDENTIFIER)).thenThrow(exception);

			var result = lookupService();

			assertThat(result).isFalse();
		}

		@DisplayName("should log warning if exception")
		@Test
		@SneakyThrows
		void shouldLogWarningIfException() {
			when(wrapper.lookupService(MESSAGE_SERVICE, READER_IDENTIFIER, AUTHOR_IDENTIFIER)).thenThrow(exception);

			lookupService();

			verify(service).logWarning(anyString(), eq(exception));
		}

		private boolean lookupService() {
			return service.lookupService(messageMetaData);
		}
	}

}