package de.ozgcloud.xta.client.core;

import static de.ozgcloud.xta.client.core.XtaExceptionHandler.*;
import static de.ozgcloud.xta.client.factory.ClientRuntimeExceptionTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.factory.ClientRuntimeExceptionTestFactory;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.PermissionDeniedException;

class XtaExceptionHandlerTest {

	@InjectMocks
	@Spy
	XtaExceptionHandler exceptionHandler;

	private XtaClientRuntimeException exception;

	@BeforeEach
	void setUp() {
		exception = ClientRuntimeExceptionTestFactory.create();
	}

	@DisplayName("derive xta client exception")
	@Nested
	class TestDeriveXtaClientException {

		@Mock
		private XtaClientException derivedClientException;

		@DisplayName("should derive xta client exception from xta client runtime exception")
		@Test
		void shouldDeriveXtaClientExceptionFromXtaClientRuntimeException() {
			doReturn(derivedClientException).when(exceptionHandler).deriveXtaClientExceptionFromClientRuntimeException(exception);

			var result = exceptionHandler.deriveXtaClientException(exception);

			assertThat(result).isEqualTo(derivedClientException);
		}

		@DisplayName("without xta client runtime exception instance")
		@Nested
		class TestWithoutXtaClientRuntimeExceptionInstance {

			@Mock
			private RuntimeException runtimeException;

			@DisplayName("should return with unknown error message")
			@Test
			void shouldReturnWithUnknownErrorMessage() {
				var result = exceptionHandler.deriveXtaClientException(runtimeException);

				assertThat(result.getMessage()).isEqualTo(UNEXPECTED_ERROR_MESSAGE);
			}

			@DisplayName("should have cause")
			@Test
			void shouldHaveCause() {
				var result = exceptionHandler.deriveXtaClientException(runtimeException);

				assertThat(result.getCause()).isEqualTo(runtimeException);
			}
		}
	}

	@DisplayName("derive xta client exception from xta client runtime exception")
	@Nested
	class TestDeriveXtaClientExceptionFromXtaClientRuntimeException {

		@DisplayName("should keep message if no cause")
		@Test
		void shouldKeepMessageIfNoCause() {
			var exceptionWithoutCause = new XtaClientRuntimeException(MESSAGE);

			var result = exceptionHandler.deriveXtaClientExceptionFromClientRuntimeException(exceptionWithoutCause);

			assertThat(result.getMessage()).isEqualTo(MESSAGE);
		}

		@DisplayName("should keep message if cause is not an xta exception")
		@Test
		void shouldKeepMessageIfCauseIsNotAnXtaException() {
			var cause = new Exception();
			var exceptionWithNoXtaCause = new XtaClientRuntimeException(MESSAGE, cause);

			var result = exceptionHandler.deriveXtaClientExceptionFromClientRuntimeException(exceptionWithNoXtaCause);

			assertThat(result.getMessage()).isEqualTo(MESSAGE);

		}

		@DisplayName("should extend detail message if cause is an xta exception")
		@Test
		void shouldExtendDetailMessageIfCauseIsAnXtaException() {
			var result = exceptionHandler.deriveXtaClientExceptionFromClientRuntimeException(exception);

			assertThat(result.getMessage()).contains(MESSAGE, CAUSE_MESSAGE, CAUSE_CODE, CAUSE_NAME);
		}
	}

	@DisplayName("get detail lines")
	@Nested
	class TestGetDetailLines {

		@DisplayName("should return empty stream if cause is null")
		@Test
		void shouldReturnEmptyStreamIfCauseIsNull() {
			var result = exceptionHandler.getDetailLines(null).toList();

			assertThat(result).isEmpty();
		}

		@DisplayName("should return empty stream if cause is not an exception")
		@Test
		void shouldReturnEmptyStreamIfCauseIsNotAnException() {
			var result = exceptionHandler.getDetailLines(new Throwable()).toList();

			assertThat(result).isEmpty();
		}

		@DisplayName("should return empty stream if cause is not an xta exception")
		@Test
		void shouldReturnEmptyStreamIfCauseIsNotAnXtaException() {
			var result = exceptionHandler.getDetailLines(new Exception()).toList();

			assertThat(result).isEmpty();
		}

		@DisplayName("should return detail lines if cause is an xta exception")
		@Test
		void shouldReturnDetailLinesIfCauseIsAnXtaException() {
			var causeException = (PermissionDeniedException) exception.getCause();
			doReturn(Stream.of(CAUSE_MESSAGE)).when(exceptionHandler).deriveDetailLinesFromException(causeException);

			var result = exceptionHandler.getDetailLines(causeException).toList();

			assertThat(result).containsExactly(CAUSE_MESSAGE);
		}
	}

	@DisplayName("derive detail lines from exception")
	@Nested
	class TestDeriveDetailLinesFromException {

		@DisplayName("should return empty stream if exception type is not present")
		@Test
		void shouldReturnEmptyStreamIfExceptionTypeIsNotPresent() {
			var result = exceptionHandler.deriveDetailLinesFromException(new Exception()).toList();

			assertThat(result).isEmpty();
		}

		@DisplayName("should return stream with error code and name if exception type is present")
		@Test
		void shouldReturnStreamWithErrorCodeAndNameIfExceptionTypeIsPresent() {
			var causeException = (PermissionDeniedException) exception.getCause();

			var result = exceptionHandler.deriveDetailLinesFromException(causeException).toList();

			assertThat(result).containsExactly(CAUSE_MESSAGE, "[%s] %s".formatted(CAUSE_CODE, CAUSE_NAME));
		}
	}

}