package de.ozgcloud.xta.client.core;

import static de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory.*;
import static de.ozgcloud.xta.client.core.XtaTLSClientParametersFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.security.KeyStore;
import java.security.KeyStoreException;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;

import org.apache.hc.core5.ssl.SSLContextBuilder;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.xta.client.factory.XtaClientConfigTestFactory;
import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.exception.XtaClientInitializationException;
import lombok.SneakyThrows;

class XtaTLSClientParametersFactoryTest {

	@Mock
	private XtaClientConfig config;

	@InjectMocks
	@Spy
	private XtaTLSClientParametersFactory factory;

	@DisplayName("create TLS client parameters")
	@Nested
	class TestCreateTlsClientParameters {
		@Mock
		SSLSocketFactory sslSocketFactory;

		@Mock
		SSLContext sslContext;

		@BeforeEach
		@SneakyThrows
		void mock() {
			doReturn(sslContext).when(factory).createXtaSslContext();
			when(sslContext.getSocketFactory()).thenReturn(sslSocketFactory);
		}

		@DisplayName("should have SSL socket factory")
		@Test
		@SneakyThrows
		void shouldHaveSslSocketFactory() {
			var tlsClientParameters = factory.create();

			assertThat(tlsClientParameters.getSSLSocketFactory()).isEqualTo(sslSocketFactory);
		}
	}

	@DisplayName("create XTA SSL context")
	@Nested
	class TestCreateXtaSslContext {

		private static final char[] PASSWORD = "password".toCharArray();

		@Mock
		XtaClientConfig.KeyStore configKeystore;

		@Mock
		KeyStore keystore;

		@Mock
		SSLContextBuilder sslContextBuilder;

		@Mock
		SSLContext sslContext;

		@BeforeEach
		@SneakyThrows
		void mock() {
			doReturn(sslContextBuilder).when(factory).createSSLContextBuilder();
			when(sslContextBuilder.build()).thenReturn(sslContext);
		}

		@DisplayName("with client key store")
		@Nested
		class TestWithClientKeyStore {

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(config.getClientCertKeystore()).thenReturn(configKeystore);
				when(configKeystore.getPassword()).thenReturn(PASSWORD);
				doReturn(keystore).when(factory).loadStore(configKeystore);
			}

			@DisplayName("should call loadKeyMaterial")
			@Test
			@SneakyThrows
			void shouldCallLoadKeyMaterial() {
				factory.createXtaSslContext();

				verify(sslContextBuilder).loadKeyMaterial(keystore, PASSWORD);
			}

			@DisplayName("should return SSL context")
			@Test
			@SneakyThrows
			void shouldSetProtocol() {
				var sslContextResult = factory.createXtaSslContext();

				assertThat(sslContextResult).isEqualTo(sslContext);
			}

			@DisplayName("should throw client initialization exception for keystore exception")
			@Test
			@SneakyThrows
			void shouldThrowClientInitializationExceptionForKeyStoreException() {
				reset(sslContextBuilder);
				doThrow(new KeyStoreException("something")).when(sslContextBuilder).loadKeyMaterial(any(), any());

				assertThatThrownBy(() -> factory.createXtaSslContext())
						.isInstanceOf(XtaClientInitializationException.class);
			}
		}

		@DisplayName("without client key store")
		@Nested
		class TestWithoutClientKeyStore {

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(config.getClientCertKeystore()).thenReturn(null);
			}

			@DisplayName("should return SSL context")
			@Test
			@SneakyThrows
			void shouldSetProtocol() {
				var sslContextResult = factory.createXtaSslContext();

				assertThat(sslContextResult).isEqualTo(sslContext);
			}
		}

		@DisplayName("with trust store")
		@Nested
		class TestWithTrustStore {
			@BeforeEach
			@SneakyThrows
			void mock() {
				when(config.getTrustStore()).thenReturn(configKeystore);
				doReturn(keystore).when(factory).loadStore(configKeystore);
			}

			@DisplayName("should call loadTrustMaterial")
			@Test
			@SneakyThrows
			void shouldCallLoadTrustMaterial() {
				factory.createXtaSslContext();

				verify(sslContextBuilder).loadTrustMaterial(keystore, null);
			}

			@DisplayName("should return SSL context")
			@Test
			@SneakyThrows
			void shouldSetProtocol() {
				var sslContextResult = factory.createXtaSslContext();

				assertThat(sslContextResult).isEqualTo(sslContext);
			}
		}

		@DisplayName("without trust store")
		@Nested
		class TestWithoutTrustStore {
			@BeforeEach
			@SneakyThrows
			void mock() {
				when(config.getTrustStore()).thenReturn(null);
			}

			@DisplayName("should return SSL context")
			@Test
			@SneakyThrows
			void shouldSetProtocol() {
				var sslContextResult = factory.createXtaSslContext();

				assertThat(sslContextResult).isEqualTo(sslContext);
			}
		}
	}

	@DisplayName("create SSL context builder")
	@Nested
	class TestCreateSslContextBuilder {

		@DisplayName("should set protocol")
		@Test
		@SneakyThrows
		void shouldSetProtocol() {
			var sslContext = factory.createSSLContextBuilder().build();

			assertThat(sslContext.getProtocol()).isEqualTo(TLS_PROTOCOL);
		}
	}

	@DisplayName("load store")
	@Nested
	class TestLoadStore {

		private final XtaClientConfig.KeyStore configKeyStore = XtaClientConfigTestFactory.createWithSelfSignedClientCert().getClientCertKeystore();

		@DisplayName("should have type")
		@Test
		@SneakyThrows
		void shouldHaveType() {
			var keystore = factory.loadStore(configKeyStore);

			assertThat(keystore.getType()).isEqualTo(configKeyStore.getType());
		}

		@DisplayName("should have entry")
		@Test
		@SneakyThrows
		void shouldHaveEntry() {
			var keystore = factory.loadStore(configKeyStore);

			assertThat(keystore.isKeyEntry(KEYSTORE_ALIAS)).isTrue();
		}
	}
}