package de.ozgcloud.xta.client.extension;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;

public record PatchHunk(
		int contextStartLineNumber,
		List<HunkLine> hunkLines
) {

	public record PatchResult(
			List<String> lines,
			int lineOffset
	) {
	}

	public PatchResult apply(Iterator<String> lines, int lineOffset) {
		var patchedLines = new ArrayList<String>();
		lineOffset = skipToContextStart(lines, lineOffset, patchedLines);

		for (var hunkLine : hunkLines) {
			switch (hunkLine.operation) {
				case CONTEXT -> {
					var contextLine = lines.next();
					lineOffset++;
					if (!hunkLine.line.equals(contextLine)) {
						throw new XtaClientRuntimeException(
								"Unexpected context line! expected='" + hunkLine.line + "' actual='" + contextLine + "', lineOffset=" + lineOffset);
					}
					patchedLines.add(contextLine);
				}
				case ADD -> patchedLines.add(hunkLine.line);
				case REMOVE -> {
					var removeLine = lines.next();
					lineOffset++;
					if (!hunkLine.line.equals(removeLine)) {
						throw new XtaClientRuntimeException(
								"Unexpected remove line! expected='" + hunkLine.line + "' actual='" + removeLine + "', lineOffset=" + lineOffset);
					}
				}
			}
		}
		return new PatchResult(patchedLines, lineOffset);
	}

	private int skipToContextStart(Iterator<String> lines, int lineOffset, List<String> patchedLines) {
		for (; lineOffset < contextStartLineNumber - 1; lineOffset++) {
			if (!lines.hasNext()) {
				throw new XtaClientRuntimeException("Unexpected end of target file! lineOffset=" + lineOffset);
			}
			patchedLines.add(lines.next());
		}
		return lineOffset;
	}

	public record HunkLine(
			String line,
			PatchOperation operation
	) {
	}

	public enum PatchOperation {
		CONTEXT,
		ADD,
		REMOVE
	}

	private record HunkHeader(
			int sourceStartLineNumber,
			int sourceLineCount
	) {
	}

	public static List<PatchHunk> parseHunks(Iterator<String> lines) {
		var hunks = new ArrayList<PatchHunk>();
		while (lines.hasNext()) {
			var line = lines.next();
			if (line.startsWith("@@")) {
				hunks.add(parseHunk(parseHunkHeader(line), lines));
			}
		}
		return hunks;
	}

	private static PatchHunk parseHunk(HunkHeader header, Iterator<String> lines) {
		var hunkLines = new ArrayList<HunkLine>();
		var sourceLineOffset = 0;
		while (sourceLineOffset < header.sourceLineCount()) {
			if (!lines.hasNext()) {
				throw new XtaClientRuntimeException("Unexpected end of patch hunk! sourceLineOffset=" + sourceLineOffset + ", header=" + header);
			}
			var line = lines.next();
			var operation = switch (line.substring(0, 1)) {
				case " " -> PatchOperation.CONTEXT;
				case "+" -> PatchOperation.ADD;
				case "-" -> PatchOperation.REMOVE;
				default -> throw new XtaClientRuntimeException(
						"Unexpected line in patch hunk! line=" + line + "sourceLineOffset=" + sourceLineOffset + ", header=" + header);
			};
			sourceLineOffset += switch (operation) {
				case PatchOperation.CONTEXT, PatchOperation.REMOVE -> 1;
				case PatchOperation.ADD -> 0;
			};
			hunkLines.add(new HunkLine(line.substring(1), operation));

		}
		return new PatchHunk(header.sourceStartLineNumber(), hunkLines);
	}

	private static HunkHeader parseHunkHeader(String line) {
		var sourcePart = line.split(" ")[1].substring(1);
		var numbers = sourcePart.split(",");
		return new HunkHeader(Integer.parseInt(numbers[0]), Integer.parseInt(numbers[1]));
	}
}
