package de.ozgcloud.xta.client.extension;

import static de.ozgcloud.xta.client.factory.ZipFileTestFactory.*;
import static java.util.Collections.*;
import static java.util.stream.Collectors.*;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.regex.Pattern;

import jakarta.activation.DataHandler;
import jakarta.annotation.Nullable;
import jakarta.mail.util.ByteArrayDataSource;
import jakarta.validation.constraints.NotBlank;

import org.apache.commons.codec.Resources;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.io.IOUtils;
import org.yaml.snakeyaml.Yaml;

import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaIdentifier;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import lombok.Builder;

public class XtaMessageExampleLoader {

	static Yaml yaml = new Yaml();

	@SuppressWarnings("unchecked")
	public static XtaMessage load(MessageExampleConfig config) {
		var messageDataMap = loadXtaMessageDataMap(config.messageLabel);
		var metaDataMap = getChild(messageDataMap, "metaData");
		var metaData = mapXtaMessageMetadata(metaDataMap, config);
		var messageFileProcessor = createMessageFileProcessor(
				metaData,
				readPatches(config.messageLabel, config.patchLabel)
		);

		return XtaMessage.builder()
				.metaData(metaData)
				.messageFile(mapXtaFile(
						getChild(messageDataMap, "messageFile"),
						config.messageLabel + "/message",
						messageFileProcessor
				))
				.attachmentFiles(mapXtaFiles((List<Map<String, Object>>) messageDataMap.get("attachmentFiles"), config.messageLabel + "/attachment"))
				.build();
	}

	private static List<Patch> readPatches(String messageLabel, String patchLabel) {
		return Optional.ofNullable(patchLabel)
				.map(label -> Patch.parsePatches(
						new String(
								readBytesFromResource(
										getMessageResourcePath(messageLabel, patchLabel)
								),
								StandardCharsets.UTF_8)
				))
				.orElse(emptyList());
	}

	private static MessageFileProcessor createMessageFileProcessor(XtaMessageMetaData metaData, List<Patch> patches) {
		var templateValues = Map.of(
				"authorIdentifier.prefix", identifierPart(metaData.authorIdentifier(), 0),
				"authorIdentifier.value", identifierPart(metaData.authorIdentifier(), 1),
				"readerIdentifier.prefix", identifierPart(metaData.readerIdentifier(), 0),
				"readerIdentifier.value", identifierPart(metaData.readerIdentifier(), 1)
		);
		return (path, content) -> {
			var patchedContent = Patch.findPatchByFilePathSuffix(patches, path)
					.map(patch -> patch.apply(new String(content, StandardCharsets.UTF_8)))
					.map(s -> s.getBytes(StandardCharsets.UTF_8))
					.orElse(content);
			if (path.endsWith(".xml")) {
				return fillTemplateString(
						new String(patchedContent, StandardCharsets.UTF_8),
						templateValues
				).getBytes(StandardCharsets.UTF_8);
			}
			return patchedContent;
		};
	}

	private static String identifierPart(XtaIdentifier identifier, int index) {
		return identifier == null ? "?" : identifier.value().split(":", 2)[index];
	}

	@Builder
	public record MessageExampleConfig(
			@NotBlank String messageLabel,
			@Nullable String patchLabel,
			@Nullable String messageId,
			@Nullable XtaIdentifier author,
			@Nullable XtaIdentifier reader) {
	}

	@SuppressWarnings("unchecked")
	private static Map<String, Object> getChild(Map<String, Object> parent, String key) {
		return Objects.requireNonNull((Map<String, Object>) MapUtils.getMap(parent, key), "Missing key: %s".formatted(key));
	}

	private static XtaFile mapXtaFile(
			Map<String, Object> messageFile,
			String resourcePrefix,
			MessageFileProcessor messageFileProcessor) {
		Function<String, String> getString = key -> Objects.requireNonNull(
				MapUtils.getString(messageFile, key), "[Failed mapping for %s] Missing key: %s".formatted(resourcePrefix, key));

		var name = getString.apply("name");
		var path = getMessageResourcePath(resourcePrefix, name);

		var data = name.endsWith(".zip")
				? createAttachmentZipFile(path, messageFileProcessor)
				: messageFileProcessor.process(path, readBytesFromResource(path));
		return XtaFile.builder()
				.name(name)
				.contentType(getString.apply("contentType"))
				.content(createContentDataHandler(data))
				.id(MapUtils.getString(messageFile, "id"))
				.build();
	}

	private static List<XtaFile> mapXtaFiles(List<Map<String, Object>> attachmentFiles, String resourcePrefix) {
		return attachmentFiles.stream()
				.map(messageFileMap -> mapXtaFile(messageFileMap, resourcePrefix, (path, content) -> content))
				.toList();
	}

	private static XtaMessageMetaData mapXtaMessageMetadata(Map<String, Object> metaData, MessageExampleConfig config) {
		Function<String, String> getString = key -> Objects.requireNonNull(
				MapUtils.getString(metaData, key), "metaData key missing: %s".formatted(key));
		return XtaMessageMetaData.builder()
				.service(getString.apply("service"))
				.businessScenarioCode(getString.apply("businessScenarioCode"))
				.businessScenarioListUri(getString.apply("businessScenarioListUri"))
				.businessScenarioListVersionId(getString.apply("businessScenarioListVersionId"))
				.messageTypeCode(getString.apply("messageTypeCode"))
				.messageTypePayloadSchema(getString.apply("messageTypePayloadSchema"))
				.messageId(getIfConfigNull(
						config.messageId, () -> MapUtils.getString(metaData, "messageId")))
				.authorIdentifier(getIfConfigNull(
						config.author,
						() -> mapIdentifier(getChild(metaData, "authorIdentifier"))))
				.readerIdentifier(getIfConfigNull(
						config.reader,
						() -> mapIdentifier(getChild(metaData, "readerIdentifier"))))
				.build();
	}

	private static <T> T getIfConfigNull(T configValue, Supplier<T> supplier) {
		return configValue != null ? configValue : supplier.get();
	}

	private static XtaIdentifier mapIdentifier(Map<String, Object> identifier) {
		Function<String, String> getString = key -> (String) identifier.get(key);
		return XtaIdentifier.builder()
				.name(getString.apply("name"))
				.value(getString.apply("value"))
				.category(getString.apply("category"))
				.build();
	}

	private static LinkedHashMap<String, Object> loadXtaMessageDataMap(String messageLabel) {
		var yamlFileData = readBytesFromResource(getMessageResourcePath(messageLabel, "xta-message.yaml"));
		return yaml.load(new String(yamlFileData));
	}

	private static String getMessageResourcePath(String prefix, String suffix) {
		return "messages/%s/%s".formatted(prefix, suffix);
	}

	private static DataHandler createContentDataHandler(byte[] data) {
		return new DataHandler(new ByteArrayDataSource(data, "application/octet-stream"));
	}

	private static byte[] readBytesFromResource(String path) {
		try {
			return IOUtils.toByteArray(Objects.requireNonNull(Resources.getInputStream(path), "Resource not found: %s".formatted(path)));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static String fillTemplateString(final String templateString, Map<String, String> templateValues) {
		var searchPattern = createPatternMatchingAny(templateValues.keySet());
		var builder = new StringBuilder();
		int start = 0;
		for (var match : searchPattern.matcher(templateString).results().toList()) {
			var templateKey = match.group(1);
			builder.append(templateString, start, match.start());
			builder.append(templateValues.getOrDefault(templateKey, ""));
			start = match.end();
		}
		builder.append(templateString, start, templateString.length());
		return builder.toString();
	}

	private static Pattern createPatternMatchingAny(Collection<String> needles) {
		return Pattern.compile(String.join("",
				Pattern.quote("${"),
				"(",
				needles.stream()
						.map(Pattern::quote)
						.collect(joining("|")),
				")",
				Pattern.quote("}")
		));
	}

}
