package de.ozgcloud.xta.client.factory;

import java.io.ByteArrayOutputStream;
import java.math.BigInteger;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.security.cert.Certificate;
import java.util.Date;
import java.util.List;

import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;

import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.model.XtaIdentifier;
import lombok.SneakyThrows;

public class XtaClientConfigTestFactory {

	public static final String SELF_IDENTIFIER_VALUE = "dp:selfIdentifier";
	public static final XtaIdentifier SELF_IDENTIFIER = XtaIdentifier.builder()
			.value(SELF_IDENTIFIER_VALUE)
			.category("Generic category")
			.name("Generic Name")
			.build();
	public static final int MAX_LIST_ITEMS = 10;
	public static final XtaIdentifier SELF_IDENTIFIER2 = XtaIdentifier.builder()
			.value("dp:selfIdentifier2")
			.category("Generic category")
			.name("Generic Name")
			.build();

	public static final String KEYSTORE_ALIAS = "selfSigned";

	static final String MANAGEMENT_PORT_SVC = "https://../managementPort.svc";
	static final String SEND_PORT_SVC = "https://../sendPort.svc";
	static final String MSG_BOX_PORT_SVC = "https://../msgBoxPort.svc";

	public static XtaClientConfig create() {
		return createBuilder().build();
	}

	public static XtaClientConfig createWithSelfSignedClientCert() {
		return createBuilder()
				.clientCertKeystore(createClientCertKeyStore())
				.build();
	}

	public static XtaClientConfig.XtaClientConfigBuilder createBuilder() {
		return XtaClientConfig.builder()
				.managementServiceUrl(MANAGEMENT_PORT_SVC)
				.sendServiceUrl(SEND_PORT_SVC)
				.msgBoxServiceUrl(MSG_BOX_PORT_SVC)
				.clientIdentifiers(List.of(SELF_IDENTIFIER, SELF_IDENTIFIER2))
				.maxListItems(MAX_LIST_ITEMS)
				.logSoapRequests(true)
				.logSoapResponses(true)
				.schemaValidation(true);
	}

	@SneakyThrows
	private static XtaClientConfig.KeyStore createClientCertKeyStore() {
		var password = "password".toCharArray();
		var type = KeyStore.getDefaultType();

		// Step 1: Create a KeyStore
		var keyStore = KeyStore.getInstance(type);
		keyStore.load(null, null); // Initialize KeyStore

		// Step 2: Generate a Key Pair
		var keyPairGenerator = KeyPairGenerator.getInstance("EC");
		keyPairGenerator.initialize(384);
		var keyPair = keyPairGenerator.generateKeyPair();

		// Step 3: Create a Self-Signed Certificate
		var issuerName = new X500Name("CN=Test, OU=Development, O=MyCompany, L=MyCity, ST=MyState, C=US");
		var serialNumber = new BigInteger(64, new SecureRandom());
		var startDate = new Date();
		var endDate = new Date(startDate.getTime() + 365L * 24L * 60L * 60L * 1000L); // 1 year validity

		var contentSigner = new JcaContentSignerBuilder("SHA256withECDSA").build(keyPair.getPrivate());
		var certBuilder = new JcaX509v3CertificateBuilder(
				issuerName, serialNumber, startDate, endDate, issuerName, keyPair.getPublic());

		var certHolder = certBuilder.build(contentSigner);
		var cert = new JcaX509CertificateConverter().setProvider(new BouncyCastleProvider()).getCertificate(certHolder);

		// Step 4: Store the Key Pair and Certificate in the KeyStore
		keyStore.setKeyEntry(KEYSTORE_ALIAS, keyPair.getPrivate(), password, new Certificate[] { cert });

		try (var os = new ByteArrayOutputStream()) {
			keyStore.store(os, password);
			return XtaClientConfig.KeyStore.builder()
					.content(os.toByteArray())
					.type(type)
					.password(password)
					.build();
		}
	}

	public static XtaClientConfig.KeyStore.KeyStoreBuilder createKeystoreBuilder() {
		return XtaClientConfig.KeyStore.builder()
				.content(new byte[] { 0, 1 })
				.type("AAAA")
				.password("".toCharArray());
	}
}
