package de.ozgcloud.xta.client.xdomea;

import static de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.client.factory.XdomeaXmlValuesTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaFileTestFactory.*;
import static de.ozgcloud.xta.client.xdomea.XdomeaMetaDataValidator.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;

import jakarta.activation.DataHandler;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.factory.XdomeaXmlValuesTestFactory;
import de.ozgcloud.xta.client.factory.XtaFileTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaXmlValues;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReader;
import lombok.SneakyThrows;

class XdomeaMetaDataValidatorTest {

	@Spy
	@InjectMocks
	private XdomeaMetaDataValidator validator;

	@Mock
	private ZipFileEntryReader zipFileEntryReader;

	@DisplayName("validate")
	@Nested
	class TestValidate {

		private XtaFile xdomeaZipFile;

		private XdomeaXmlValues xdomeaXmlValues;

		private XtaMessageMetaData messageMetadataData;

		@BeforeEach
		@SneakyThrows
		void mock() {
			xdomeaZipFile = XtaFileTestFactory.create();
			xdomeaXmlValues = XdomeaXmlValuesTestFactory.create();
			messageMetadataData = XtaMessageMetaDataTestFactory.create();
			doNothing().when(validator).validateIdentifierPrefixes(any(), any());
			doNothing().when(validator).validateZipFileName(any(), any(), any());
			doNothing().when(validator).validatePrimaryDocumentReferences(any(), any());
		}

		@DisplayName("should verify identifier prefixes")
		@Test
		@SneakyThrows
		void shouldVerifyIdentifierPrefixes() {
			validate();

			verify(validator).validateIdentifierPrefixes(
					AUTHOR_ID_PREFIX,
					READER_ID_PREFIX
			);
		}

		@DisplayName("should verify zip file name")
		@Test
		@SneakyThrows
		void shouldVerifyZipFileName() {
			validate();

			verify(validator).validateZipFileName(
					XTA_FILE_NAME,
					PROCESS_ID,
					MESSAGE_TYPE_CODE
			);
		}

		@DisplayName("should verify primary document references")
		@Test
		@SneakyThrows
		void shouldVerifyPrimaryDocumentReferences() {
			validate();

			verify(validator).validatePrimaryDocumentReferences(
					xdomeaZipFile,
					PRIMARY_DOCUMENT_NAMES
			);
		}

		@SneakyThrows
		private void validate() {
			validator.validate(xdomeaZipFile, xdomeaXmlValues, messageMetadataData);
		}
	}

	@DisplayName("validate identifier prefixes")
	@Nested
	class TestValidateIdentifierPrefixes {

		@BeforeEach
		@SneakyThrows
		void mock() {
			doNothing().when(validator).validateIdentifierPrefix(any(), any(), any());
		}

		@DisplayName("should validate author identifier prefix")
		@Test
		@SneakyThrows
		void shouldValidateAuthorIdentifierPrefix() {
			validator.validateIdentifierPrefixes(AUTHOR_ID_PREFIX, READER_ID_PREFIX);

			verify(validator).validateIdentifierPrefix(
					AUTHOR_ID_PREFIX,
					MetadataMapper.AUTHOR_ID_PREFIX,
					AUTHOR_ID_PREFIX_NAME
			);
		}

		@DisplayName("should validate reader identifier prefix")
		@Test
		@SneakyThrows
		void shouldValidateReaderIdentifierPrefix() {
			validator.validateIdentifierPrefixes(AUTHOR_ID_PREFIX, READER_ID_PREFIX);

			verify(validator).validateIdentifierPrefix(
					READER_ID_PREFIX,
					MetadataMapper.READER_ID_PREFIX,
					READER_ID_PREFIX_NAME
			);
		}
	}

	@DisplayName("validate identifier prefix")
	@Nested
	class TestValidateIdentifierPrefix {

		private static final String EXPECTED_PREFIX = "prefix";
		private static final String PREFIX_NAME = "prefixName";

		@DisplayName("should normalize prefix")
		@ParameterizedTest
		@ValueSource(strings = {
				"prefix:",
				"prefix",
		})
		@SneakyThrows
		void shouldNormalizePrefix(String prefix) {
			validator.validateIdentifierPrefix(prefix, EXPECTED_PREFIX, PREFIX_NAME);
		}

		@DisplayName("should throw client exception for wrong prefix")
		@ParameterizedTest
		@ValueSource(strings = {
				"prefi",
				"prefix::",
				"prefix$:",
				"prefix$",
		})
		void shouldThrowClientExceptionForWrongPrefix(String prefix) {
			assertThatThrownBy(() -> validator.validateIdentifierPrefix(prefix, EXPECTED_PREFIX, PREFIX_NAME))
					.isInstanceOf(XtaClientException.class)
					.hasMessageContaining(PREFIX_NAME);
		}

	}

	@DisplayName("validate zip file name")
	@Nested
	class TestValidateZipFileName {

		private static final String PROCESS_ID = "processId";
		private static final String MESSAGE_TYPE_CODE = "messageTypeCode";
		private static final String VALID_ZIP_FILE_NAME = "processId_messageTypeCode.zip";

		@DisplayName("should not throw exception for valid zip file name")
		@Test
		void shouldNotThrowExceptionForValidZipFileName() {
			assertThatCode(() -> validator.validateZipFileName(VALID_ZIP_FILE_NAME, PROCESS_ID, MESSAGE_TYPE_CODE))
					.doesNotThrowAnyException();
		}

		@DisplayName("should throw client exception for wrong zip file name")
		@ParameterizedTest
		@ValueSource(strings = {
				"$processId_messageTypeCode.zip",
				"processId_me$ssageTypeCode.zip",
				"processId_messageTypeCode.zip$",
		})
		void shouldThrowClientExceptionForWrongZipFileName(String wrongZipFileName) {
			assertThatThrownBy(() -> validator.validateZipFileName(wrongZipFileName, PROCESS_ID, MESSAGE_TYPE_CODE))
					.isInstanceOf(XtaClientException.class);
		}

	}

	@DisplayName("validate primary document references")
	@Nested
	class TestValidatePrimaryDocumentReferences {

		private static final List<String> VALID_ZIP_ENTRY_NAMES = List.of("doc.pdf", "doc.xml", "other.doc");
		private static final List<String> INVALID_ZIP_ENTRY_NAMES = List.of("doc.pdf");

		@Mock
		private XtaFile xdomeaZipFile;

		@Mock
		private DataHandler contentDataHandler;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xdomeaZipFile.content()).thenReturn(contentDataHandler);

		}

		@DisplayName("should not throw exception for valid primary document references")
		@Test
		@SneakyThrows
		void shouldNotThrowExceptionForValidPrimaryDocumentReferences() {
			when(zipFileEntryReader.getEntryNames(contentDataHandler)).thenReturn(VALID_ZIP_ENTRY_NAMES);

			validator.validatePrimaryDocumentReferences(xdomeaZipFile, PRIMARY_DOCUMENT_NAMES);
		}

		@DisplayName("should throw client exception for missing primary document references")
		@Test
		@SneakyThrows
		void shouldThrowClientExceptionForMissingPrimaryDocumentReferences() {
			when(zipFileEntryReader.getEntryNames(contentDataHandler)).thenReturn(INVALID_ZIP_ENTRY_NAMES);

			assertThatThrownBy(() -> validator.validatePrimaryDocumentReferences(xdomeaZipFile, PRIMARY_DOCUMENT_NAMES))
					.isInstanceOf(XtaClientException.class);
		}
	}

}