package de.ozgcloud.xta.client.xdomea;

import static de.ozgcloud.xta.client.xdomea.XdomeaXtaMessageCreatorFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.net.URL;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSResourceResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;

import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaValueReader;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaValueReaderFactory;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReader;
import lombok.SneakyThrows;

class XdomeaXtaMessageCreatorFactoryTest {

	@InjectMocks
	@Spy
	private XdomeaXtaMessageCreatorFactory factory;

	@Mock
	private XdomeaValueReaderFactory xdomeaValueReaderFactory;

	@Mock
	private ZipFileEntryReader zipFileEntryReader;

	@Mock
	private DocumentBuilderFactory documentBuilderFactory;

	@Mock
	private MetadataMapper metadataMapper;

	@Mock
	private XdomeaMetaDataValidatorFactory xdomeaMetaDataValidatorFactory;

	@Mock
	private SchemaFactory schemaFactory;

	@Mock
	private ClassLoader classLoader;

	@Mock
	private ErrorHandler errorHandler;

	@DisplayName("create")
	@Nested
	class TestCreate {

		@Mock
		private XdomeaValueReader xdomeaValueReader;

		@Mock
		private DocumentBuilder documentBuilder;

		@Mock
		private ParserConfigurationException parserConfigurationException;

		@Mock
		private XdomeaMetaDataValidator metaDataValidator;

		@Mock
		private Schema schema;

		@SneakyThrows
		private void setupMocks() {
			when(documentBuilderFactory.newDocumentBuilder()).thenReturn(documentBuilder);
			doReturn(schema).when(factory).createSchema();
			when(xdomeaValueReaderFactory.create()).thenReturn(xdomeaValueReader);
			when(xdomeaMetaDataValidatorFactory.create()).thenReturn(metaDataValidator);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			setupMocks();

			var result = factory.create();

			assertThat(result).usingRecursiveComparison().isEqualTo(XdomeaXtaMessageCreator.builder()
					.documentBuilder(documentBuilder)
					.xdomeaValueReader(xdomeaValueReader)
					.zipFileEntryReader(zipFileEntryReader)
					.metadataMapper(metadataMapper)
					.metaDataValidator(metaDataValidator)
					.build());
		}

		@DisplayName("should disallow doctype declarations")
		@Test
		@SneakyThrows
		void shouldDisallowDoctypeDeclarations() {
			setupMocks();

			factory.create();

			verify(documentBuilderFactory).setFeature(DISALLOW_DOCTYPE_DECL_FEATURE_URL, true);
		}

		@DisplayName("should set document builder namespace aware")
		@Test
		@SneakyThrows
		void shouldSetDocumentBuilderNamespaceAware() {
			setupMocks();

			factory.create();

			verify(documentBuilderFactory).setNamespaceAware(true);
		}

		@DisplayName("should set document builder schema")
		@Test
		@SneakyThrows
		void shouldSetDocumentBuilderSchema() {
			setupMocks();

			factory.create();

			verify(documentBuilderFactory).setSchema(schema);
		}

		@DisplayName("should set document builder error handler")
		@Test
		@SneakyThrows
		void shouldSetDocumentBuilderErrorHandler() {
			setupMocks();

			factory.create();

			verify(documentBuilder).setErrorHandler(errorHandler);
		}

		@DisplayName("should throw technical initialization exception")
		@Test
		@SneakyThrows
		void shouldThrowTechnicalInitializationException() {
			doThrow(parserConfigurationException).when(documentBuilderFactory).setFeature(DISALLOW_DOCTYPE_DECL_FEATURE_URL, true);

			assertThatThrownBy(() -> factory.create())
					.isInstanceOf(XtaClientRuntimeException.class)
					.hasCause(parserConfigurationException);
		}

	}

	@DisplayName("create schema")
	@Nested
	class TestCreateSchema {

		@Mock
		private Schema schema;

		@Mock
		private LSResourceResolver resolver;

		@Mock
		StreamSource streamSource1;

		@Mock
		StreamSource streamSource2;

		private StreamSource[] schemaSources;

		@BeforeEach
		void mock() {
			doReturn(resolver).when(factory).createResourceResolver();
			schemaSources = new StreamSource[] { streamSource1, streamSource2 };
			doReturn(schemaSources).when(factory).createSchemaStreamSources();
		}

		@SneakyThrows
		private void setupSuccess() {
			when(schemaFactory.newSchema(schemaSources)).thenReturn(schema);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			setupSuccess();

			var result = factory.createSchema();

			assertThat(result).isEqualTo(schema);
		}

		@DisplayName("should set resource resolver")
		@Test
		@SneakyThrows
		void shouldSetResourceResolver() {
			setupSuccess();

			factory.createSchema();

			verify(schemaFactory).setResourceResolver(resolver);
		}

		@DisplayName("should throw runtime initialization exception")
		@Test
		@SneakyThrows
		void shouldThrowTechnicalInitializationException() {
			doThrow(new SAXException()).when(schemaFactory).newSchema(schemaSources);

			assertThatThrownBy(() -> factory.createSchema())
					.isInstanceOf(XtaClientRuntimeException.class);
		}
	}

	@DisplayName("create schema stream sources")
	@Nested
	class TestCreateSchemaStreamSources {

		private final Map<String, StreamSource> STREAM_SOURCE_MAP = SCHEMA_RESOURCES.stream()
				.collect(Collectors.toMap(
						resourcePath -> resourcePath,
						(r) -> mock(StreamSource.class),
						(key1, key2) -> key1,
						LinkedHashMap::new
				));

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			doAnswer(invocation -> STREAM_SOURCE_MAP.get((String) invocation.getArgument(0)))
					.when(factory).createSchemaStreamSource(anyString());

			var result = factory.createSchemaStreamSources();

			assertThat(result).isEqualTo(STREAM_SOURCE_MAP.values().toArray(new StreamSource[0]));
		}
	}

	@DisplayName("create schema stream source")
	@Nested
	class TestCreateSchemaStreamSource {

		@Mock
		private URL resourceUrl;

		@DisplayName("should return with systemId")
		@Test
		void shouldReturnWithSystemId() {
			var systemId = "exampleSystemId";
			var resourcePath = "exampleResourcePath";
			when(classLoader.getResource(resourcePath)).thenReturn(resourceUrl);
			when(resourceUrl.toExternalForm()).thenReturn(systemId);

			var streamSource = factory.createSchemaStreamSource(resourcePath);

			assertThat(streamSource.getSystemId()).isEqualTo(systemId);
		}

	}

	@DisplayName("create resource resolver")
	@Nested
	class TestCreateResourceResolver {

		private final String publicId = "examplePublicId";

		@Mock
		InputStream inputStream;

		@DisplayName("should return handler which returns cached schema resource")
		@ParameterizedTest
		@MethodSource("createCachedSchemaResourcesArguments")
		void shouldReturnHandlerWhichReturnsCachedSchemaResource(String systemId, String resourceName) {
			when(classLoader.getResourceAsStream(resourceName)).thenReturn(inputStream);

			var resource = createLSInput(systemId);

			assertThat(resource.getByteStream()).isEqualTo(inputStream);
		}

		@DisplayName("should return handler which sets publicId")
		@ParameterizedTest
		@MethodSource("createCachedSchemaResourcesArguments")
		void shouldReturnHandlerWhichSetsPublicId(String systemId, String resourceName) {
			when(classLoader.getResourceAsStream(resourceName)).thenReturn(inputStream);

			var resource = createLSInput(systemId);

			assertThat(resource.getPublicId()).isEqualTo(publicId);
		}

		static Stream<Arguments> createCachedSchemaResourcesArguments() {
			return CACHED_SCHEMA_RESOURCES.entrySet().stream()
					.map(entry -> Arguments.of(entry.getKey(), entry.getValue()));
		}

		@DisplayName("should return handler which returns null for unknown systemIds")
		@Test
		void shouldReturnHandlerWhichReturnsNullForUnknownSystemIds() {
			var resource = createLSInput("unknownSystemId");

			assertThat(resource).isNull();
		}

		private LSInput createLSInput(String systemId) {
			return factory.createResourceResolver()
					.resolveResource("?", "?", publicId, systemId, "?");
		}
	}

}