package de.ozgcloud.xta.client.xdomea;

import static de.ozgcloud.xta.client.extension.XtaServerSetupExtensionTestUtil.*;
import static de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper.*;
import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;

import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.extension.XtaMessageExampleLoader;
import de.ozgcloud.xta.client.model.XtaFile;
import lombok.SneakyThrows;

class XdomeaXtaMessageCreatorITCase {

	private XdomeaXtaMessageCreator creator;

	@BeforeEach
	void setup() {
		creator = XdomeaXtaMessageCreator.createInstance();
	}

	@DisplayName("create message")
	@Nested
	class TestCreateMessage {

		@DisplayName("should use reader")
		@SneakyThrows
		@ParameterizedTest
		@ValueSource(strings = { "dfoerdermittel", "abgabe0401-kleiner-waffenschein" })
		void shouldUseReader(String messageLabel) {
			var xdomeaZipFile = loadMessageFile(messageLabel);

			var xtaMessage = creator.createMessage(xdomeaZipFile);

			assertThat(xtaMessage.metaData().readerIdentifier().value()).isEqualTo(READER_CLIENT_IDENTIFIER1.value());
		}

		@DisplayName("should use author")
		@SneakyThrows
		@ParameterizedTest
		@ValueSource(strings = { "dfoerdermittel", "abgabe0401-kleiner-waffenschein" })
		void shouldUseAuthor(String messageLabel) {
			var xdomeaZipFile = loadMessageFile(messageLabel);

			var xtaMessage = creator.createMessage(xdomeaZipFile);

			assertThat(xtaMessage.metaData().authorIdentifier().value()).isEqualTo(AUTHOR_CLIENT_IDENTIFIER.value());
		}

		@DisplayName("should use message type code for 0201")
		@Test
		@SneakyThrows
		void shouldUseMessageTypeCodeFor0201() {
			var xdomeaZipFile = loadMessageFile("dfoerdermittel");

			var xtaMessage = creator.createMessage(xdomeaZipFile);

			assertThat(xtaMessage.metaData().messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE_0201);
		}

		@DisplayName("should throw client exception for invalid message 0201")
		@ParameterizedTest
		@ValueSource(strings = {
				"invalid-behoerdenkennung-order.patch",
				"invalid-Kennung-missing.patch",
		})
		void shouldThrowClientExceptionForInvalidMessage0201(String patchName) {
			var invalidMessageZipFile = loadMessageFileWithPatch(
					"dfoerdermittel",
					patchName
			);

			assertThatThrownBy(() -> creator.createMessage(invalidMessageZipFile))
					.isInstanceOf(XtaClientException.class);
		}

		@DisplayName("should not throw with valid message 0201")
		@Test
		void shouldNotThrowWithValidMessage0201() {
			var validMessageZipFile = loadMessageFileWithPatch(
					"dfoerdermittel",
					"valid-behoerdenkennung-name.patch"
			);

			assertThatCode(() -> creator.createMessage(validMessageZipFile))
					.doesNotThrowAnyException();
		}

		@DisplayName("should use message type code for 0401")
		@Test
		@SneakyThrows
		void shouldUseMessageTypeCodeFor0401() {
			var xdomeaZipFile = loadMessageFile("abgabe0401-kleiner-waffenschein");

			var xtaMessage = creator.createMessage(xdomeaZipFile);

			assertThat(xtaMessage.metaData().messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE_0401);
		}

		@DisplayName("should throw client exception for invalid message 0401")
		@ParameterizedTest
		@ValueSource(strings = {
				"invalid-ProzessID.patch",
				"invalid-reference-to-primary-document.patch",
				"invalid-Praefix-author.patch",
				"invalid-Praefix-reader.patch",
		})
		void shouldThrowClientExceptionForInvalidMessage0401(String patchName) {
			var invalidMessageZipFile = loadMessageFileWithPatch(
					"abgabe0401-kleiner-waffenschein",
					patchName
			);

			assertThatThrownBy(() -> creator.createMessage(invalidMessageZipFile))
					.isInstanceOf(XtaClientException.class);
		}

		@DisplayName("should not throw with valid message 0401")
		@ParameterizedTest
		@ValueSource(strings = {
				"valid-unknown-namespaces.patch",
				"valid-ignored-schemaLocation.patch"
		})
		void shouldNotThrowWithValidMessage0401(String patchName) {
			var validMessageZipFile = loadMessageFileWithPatch(
					"abgabe0401-kleiner-waffenschein",
					patchName
			);

			assertThatCode(() -> creator.createMessage(validMessageZipFile))
					.doesNotThrowAnyException();
		}

	}

	private XtaMessageExampleLoader.MessageExampleConfig.MessageExampleConfigBuilder createExampleConfigBuilder() {
		return XtaMessageExampleLoader.MessageExampleConfig.builder()
				.author(AUTHOR_CLIENT_IDENTIFIER)
				.reader(READER_CLIENT_IDENTIFIER1);
	}

	private XtaFile loadMessageFileWithPatch(String messageLabel, String patchLabel) {
		return XtaMessageExampleLoader.load(createExampleConfigBuilder()
				.messageLabel(messageLabel)
				.patchLabel(patchLabel)
				.build()).messageFile();
	}

	private XtaFile loadMessageFile(String messageLabel) {
		return loadMessageFileWithPatch(messageLabel, null);
	}

}