package de.ozgcloud.xta.client.xdomea;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.function.Function;

import javax.xml.parsers.DocumentBuilder;

import jakarta.activation.DataHandler;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.w3c.dom.Document;

import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.factory.XdomeaXmlValuesTestFactory;
import de.ozgcloud.xta.client.factory.XtaFileTestFactory;
import de.ozgcloud.xta.client.factory.XtaMessageMetaDataTestFactory;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaValueReader;
import de.ozgcloud.xta.client.xdomea.reader.XdomeaXmlValues;
import de.ozgcloud.xta.client.xdomea.reader.ZipFileEntryReader;
import lombok.SneakyThrows;

class XdomeaXtaMessageCreatorTest {

	@InjectMocks
	@Spy
	private XdomeaXtaMessageCreator creator;

	@Mock
	private XdomeaValueReader xdomeaValueReader;

	@Mock
	private DocumentBuilder documentBuilder;

	@Mock
	private MetadataMapper metadataMapper;

	@Mock
	private ZipFileEntryReader zipFileEntryReader;

	@Mock
	private XdomeaMetaDataValidator metaDataValidator;

	@DisplayName("create message")
	@Nested
	class TestCreateMessage {

		@Mock
		private XtaFile xdomeaZipFile;

		@Mock
		private XtaMessageMetaData metaData;

		@BeforeEach
		@SneakyThrows
		void mock() {
			doReturn(metaData).when(creator).deriveValidMetaData(xdomeaZipFile);
		}

		@DisplayName("should return metadata")
		@Test
		void shouldReturnMetadata() {
			var result = createMessage();

			assertThat(result.metaData()).isEqualTo(metaData);
		}

		@DisplayName("should return message file")
		@Test
		void shouldReturnMessageFile() {
			var result = createMessage();

			assertThat(result.messageFile()).isEqualTo(xdomeaZipFile);
		}

		@DisplayName("should return empty attachments")
		@Test
		void shouldReturnEmptyAttachments() {
			var result = createMessage();

			assertThat(result.attachmentFiles()).isEmpty();
		}

		@SneakyThrows
		XtaMessage createMessage() {
			return creator.createMessage(xdomeaZipFile);
		}
	}

	@DisplayName("derive valid meta data")
	@Nested
	class TestDeriveValidMetaData {

		private XtaFile xdomeaZipFile;

		private XdomeaXmlValues xdomeaXmlValues;

		private XtaMessageMetaData metaData;

		@Mock
		private XtaClientRuntimeException runtimeException;

		@BeforeEach
		@SneakyThrows
		void mock() {
			xdomeaZipFile = XtaFileTestFactory.create();
			xdomeaXmlValues = XdomeaXmlValuesTestFactory.create();
			metaData = XtaMessageMetaDataTestFactory.create();
			doReturn(xdomeaXmlValues).when(creator).readXdomeaXmlValues(xdomeaZipFile);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			mockValidMetadata();

			var result = creator.deriveValidMetaData(xdomeaZipFile);

			assertThat(result).isEqualTo(metaData);
		}

		@DisplayName("should validate")
		@Test
		@SneakyThrows
		void shouldValidate() {
			mockValidMetadata();

			creator.deriveValidMetaData(xdomeaZipFile);

			verify(metaDataValidator).validate(xdomeaZipFile, xdomeaXmlValues, metaData);
		}

		@SneakyThrows
		private void mockValidMetadata() {
			doReturn(metaData).when(metadataMapper).mapXtaMessageMetadata(xdomeaXmlValues);
			doNothing().when(metaDataValidator).validate(xdomeaZipFile, xdomeaXmlValues, metaData);
		}

		@DisplayName("should throw client exception")
		@Test
		void shouldThrowClientException() {
			doThrow(runtimeException).when(metadataMapper).mapXtaMessageMetadata(xdomeaXmlValues);

			assertThatThrownBy(() -> creator.deriveValidMetaData(xdomeaZipFile))
					.isInstanceOf(XtaClientException.class);
		}

	}

	@DisplayName("read xdomea xml values")
	@Nested
	class TestReadXdomeaXmlValues {

		@Mock
		private XtaFile xdomeaZipFile;

		@Mock
		private Document document;

		@Mock
		private XdomeaXmlValues xdomeaXmlValues;

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			doReturn(document).when(creator).readXdomeaXmlDocument(xdomeaZipFile);
			doReturn(xdomeaXmlValues).when(xdomeaValueReader).readValues(document);

			var result = creator.readXdomeaXmlValues(xdomeaZipFile);

			assertThat(result).isEqualTo(xdomeaXmlValues);
		}

	}

	@DisplayName("replace filename extension with xml")
	@Nested
	class TestReplaceFilenameExtensionWithXml {
		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = creator.replaceFilenameExtensionWithXml("some-name.with.extension");

			assertThat(result).isEqualTo("some-name.with.xml");
		}
	}

	@DisplayName("read xdomea xml document")
	@Nested
	class TestReadXdomeaXmlDocument {

		private static final String ZIP_FILE_NAME = "file.zip";
		private static final String XML_FILE_NAME = "file.xml";

		@Mock
		XtaFile xdomeaZipFile;

		@Mock
		DataHandler contentDataHandler;

		@Mock
		Document document;

		@Mock
		InputStream inputStream;

		@BeforeEach
		@SneakyThrows
		void mock() {
			when(xdomeaZipFile.name()).thenReturn(ZIP_FILE_NAME);
			when(xdomeaZipFile.content()).thenReturn(contentDataHandler);
			doReturn(XML_FILE_NAME).when(creator).replaceFilenameExtensionWithXml(ZIP_FILE_NAME);
			when(zipFileEntryReader.mapEntryData(eq(contentDataHandler), eq(XML_FILE_NAME), any())).thenAnswer(invocation -> {
				Function<InputStream, Document> function = invocation.getArgument(2);
				return function.apply(inputStream);
			});
			when(documentBuilder.parse(inputStream)).thenReturn(document);
		}

		@DisplayName("should return")
		@Test
		@SneakyThrows
		void shouldReturn() {
			var result = creator.readXdomeaXmlDocument(xdomeaZipFile);

			assertThat(result).isEqualTo(document);
		}
	}

}