package de.ozgcloud.xta.client.xdomea.mapper;

import static de.ozgcloud.xta.client.factory.XdomeaXmlValuesTestFactory.*;
import static de.ozgcloud.xta.client.xdomea.mapper.MetadataMapper.*;
import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.factory.XdomeaXmlValuesTestFactory;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;

class MetadataMapperTest {

	private final MetadataMapper metadataMapper = Mappers.getMapper(MetadataMapper.class);

	@DisplayName("map xta message metadata")
	@Nested
	class TestMapXtaMessageMetadata {

		@DisplayName("should map service 0401")
		@Test
		void shouldMapService0401() {
			var metadata = doMapping(CODE_0401);

			assertThat(metadata.service()).isEqualTo(SERVICE_XDOMEA_300);
		}

		@DisplayName("should map service 0402")
		@Test
		void shouldMapService0402() {
			var metadata = doMapping(CODE_0402);

			assertThat(metadata.service()).isEqualTo(SERVICE_XDOMEA_300);
		}

		@DisplayName("should map service 0201")
		@Test
		void shouldMapService0201() {
			var metadata = doMapping(CODE_0201);

			assertThat(metadata.service()).isEqualTo(SERVICE_XDOMEA_240);
		}

		@DisplayName("should map business scenario code")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401 })
		void shouldMapBusinessScenarioCode(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.businessScenarioCode()).isEqualTo(BUSINESS_SCENARIO_CODE);
		}

		@DisplayName("should map business scenario list uri")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401 })
		void shouldMapBusinessScenarioListUri(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.businessScenarioListUri()).isEqualTo(BUSINESS_SCENARIO_LIST_URI);
		}

		@DisplayName("should map business scenario list version id")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401 })
		void shouldMapBusinessScenarioListVersionId(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.businessScenarioListVersionId()).isEqualTo(BUSINESS_SCENARIO_LIST_VERSION_ID);
		}

		@DisplayName("should map message type code 0401")
		@Test
		void shouldMapMessageTypeCode0401() {
			var metadata = doMapping(CODE_0401);

			assertThat(metadata.messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE_0401);
		}

		@DisplayName("should map message type code 0402")
		@Test
		void shouldMapMessageTypeCode0402() {
			var metadata = doMapping(CODE_0402);

			assertThat(metadata.messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE_0402);
		}

		@DisplayName("should map message type code 0201")
		@Test
		void shouldMapMessageTypeCode0201() {
			var metadata = doMapping(CODE_0201);

			assertThat(metadata.messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE_0201);
		}

		@DisplayName("should map message type payload schema 0401")
		@Test
		void shouldMapMessageTypePayloadSchema0401() {
			var metadata = doMapping(CODE_0401);

			assertThat(metadata.messageTypePayloadSchema()).isEqualTo(MESSAGE_TYPE_PAYLOAD_SCHEMA_XDOMEA_300);
		}

		@DisplayName("should map message type payload schema 0402")
		@Test
		void shouldMapMessageTypePayloadSchema0402() {
			var metadata = doMapping(CODE_0402);

			assertThat(metadata.messageTypePayloadSchema()).isEqualTo(MESSAGE_TYPE_PAYLOAD_SCHEMA_XDOMEA_300);
		}

		@DisplayName("should map message type payload schema 0201")
		@Test
		void shouldMapMessageTypePayloadSchema0201() {
			var metadata = doMapping(CODE_0201);

			assertThat(metadata.messageTypePayloadSchema()).isEqualTo(MESSAGE_TYPE_PAYLOAD_SCHEMA_XDOMEA_240);
		}

		@DisplayName("should map author identifier value")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401, CODE_0402 })
		void shouldMapAuthorIdentifierValue(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.authorIdentifier().value()).isEqualTo(MetadataMapper.AUTHOR_ID_PREFIX + ":" + AUTHOR_ID_SUFFIX);
		}

		@DisplayName("should map author identifier category")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401, CODE_0402 })
		void shouldMapAuthorIdentifierCategory(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.authorIdentifier().category()).isEqualTo(AUTHOR_ID_CATEGORY);
		}

		@DisplayName("should map reader identifier value")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401, CODE_0402 })
		void shouldMapReaderIdentifierValue(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.readerIdentifier().value()).isEqualTo(MetadataMapper.READER_ID_PREFIX + ":" + READER_ID_SUFFIX);
		}

		@DisplayName("should map reader identifier category")
		@ParameterizedTest
		@ValueSource(strings = { CODE_0201, CODE_0401, CODE_0402 })
		void shouldMapReaderIdentifierCategory(String code) {
			var metadata = doMapping(code);

			assertThat(metadata.readerIdentifier().category()).isEqualTo(READER_ID_CATEGORY);
		}

		@DisplayName("should throw xta client runtime exception with unknown message type code")
		@Test
		void shouldThrowXtaClientRuntimeExceptionWithUnknownMessageTypeCode() {
			assertThatThrownBy(() -> doMapping("unknown"))
					.isInstanceOf(XtaClientRuntimeException.class);
		}

		private XtaMessageMetaData doMapping(String code) {
			return metadataMapper.mapXtaMessageMetadata(XdomeaXmlValuesTestFactory.createBuilder()
					.messageTypeCode(code)
					.build());
		}

	}

}